@@ -71,7 +71,7 @@ void calc_mean_and_sd_t(size_t ix_arr[], size_t st, size_t end, real_t_ *restric
     real_t m = 0;
     real_t s = 0;
     real_t m_prev = x[ix_arr[st]];
-    real_t_ xval;
+    real_t xval;
 
     if (missing_action == Fail)
     {
@@ -80,7 +80,7 @@ void calc_mean_and_sd_t(size_t ix_arr[], size_t st, size_t end, real_t_ *restric
         {
             xval = x[ix_arr[row]];
             m += (xval - m) / (real_t)(row - st + 1);
-            s += (xval - m) * (xval - m_prev);
+            s  = std::fma(xval - m, xval - m_prev, s);
             m_prev = m;
         }
 
@@ -103,7 +103,7 @@ void calc_mean_and_sd_t(size_t ix_arr[], size_t st, size_t end, real_t_ *restric
             {
                 cnt++;
                 m += (xval - m) / (real_t)cnt;
-                s += (xval - m) * (xval - m_prev);
+                s  = std::fma(xval - m, xval - m_prev, s);
                 m_prev = m;
             }
         }
@@ -139,7 +139,7 @@ void calc_mean_and_sd(size_t ix_arr[], size_t st, size_t end, real_t_ *restrict
     if (end - st + 1 < THRESHOLD_LONG_DOUBLE)
         calc_mean_and_sd_t<double, real_t_>(ix_arr, st, end, x, missing_action, x_sd, x_mean);
     else
-        calc_mean_and_sd_t<long double, real_t_>(ix_arr, st, end, x, missing_action, x_sd, x_mean);
+        calc_mean_and_sd_t<ldouble_safe, real_t_>(ix_arr, st, end, x, missing_action, x_sd, x_mean);
     x_sd = std::fmax(x_sd, SD_MIN);
 }
 
@@ -147,12 +147,12 @@ template <class real_t_, class mapping>
 void calc_mean_and_sd_weighted(size_t ix_arr[], size_t st, size_t end, real_t_ *restrict x, mapping &w,
                                MissingAction missing_action, double &restrict x_sd, double &restrict x_mean)
 {
-    long double cnt = 0;
-    double w_this;
-    double m = 0;
-    long double s = 0;
-    double m_prev = x[ix_arr[st]];
-    real_t_ xval;
+    ldouble_safe cnt = 0;
+    ldouble_safe w_this;
+    ldouble_safe m = 0;
+    ldouble_safe s = 0;
+    ldouble_safe m_prev = x[ix_arr[st]];
+    ldouble_safe xval;
     while (is_na_or_inf(m_prev) && st <= end)
     {
         m_prev = x[ix_arr[++st]];
@@ -165,14 +165,14 @@ void calc_mean_and_sd_weighted(size_t ix_arr[], size_t st, size_t end, real_t_ *
         {
             w_this = w[ix_arr[row]];
             cnt += w_this;
-            m += w_this * (xval - m) / cnt;
-            s += w_this * ((xval - m) * (xval - m_prev));
+            m = std::fma(w_this, (xval - m) / cnt, m);
+            s = std::fma(w_this, (xval - m) * (xval - m_prev), s);
             m_prev = m;
         }
     }
 
     x_mean = m;
-    x_sd   = std::sqrt(s / (long double)cnt);
+    x_sd   = std::sqrt((long double)s / (long double)cnt);
 }
 
 template <class real_t_, class mapping>
@@ -187,7 +187,7 @@ double calc_mean_only_weighted(size_t ix_arr[], size_t st, size_t end, real_t_ *
         {
             w_this = w[ix_arr[row]];
             cnt += w_this;
-            m += w_this * (x[ix_arr[row]] - m) / cnt;
+            m = std::fma(w_this, (x[ix_arr[row]] - m) / cnt, m);
         }
     }
 
@@ -234,7 +234,7 @@ void calc_mean_and_sd(size_t *restrict ix_arr, size_t st, size_t end, size_t col
             {
                 if (added == 0) m_prev = Xc[curr_pos];
                 m += (Xc[curr_pos] - m) / (real_t)(++added);
-                s += (Xc[curr_pos] - m) * (Xc[curr_pos] - m_prev);
+                s  = std::fma(Xc[curr_pos] - m, Xc[curr_pos] - m_prev, s);
                 m_prev = m;
             }
 
@@ -354,14 +354,14 @@ void calc_mean_and_sd_weighted(size_t *restrict ix_arr, size_t st, size_t end, s
     size_t ind_end_col = (size_t) Xc_ind[end_col];
     size_t *ptr_st = std::lower_bound(ix_arr + st, ix_arr + end + 1, (size_t)Xc_ind[st_col]);
 
-    long double cnt = 0.;
+    ldouble_safe cnt = 0.;
     for (size_t row = st; row <= end; row++)
         cnt += w[ix_arr[row]];
-    long double added = 0;
-    double m = 0;
-    long double s = 0;
-    double m_prev = 0;
-    double w_this;
+    ldouble_safe added = 0;
+    ldouble_safe m = 0;
+    ldouble_safe s = 0;
+    ldouble_safe m_prev = 0;
+    ldouble_safe w_this;
 
     for (size_t *row = ptr_st;
          row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
@@ -379,8 +379,8 @@ void calc_mean_and_sd_weighted(size_t *restrict ix_arr, size_t st, size_t end, s
                 w_this = w[*row];
                 if (added == 0) m_prev = Xc[curr_pos];
                 added += w_this;
-                m += w_this * (Xc[curr_pos] - m) / added;
-                s += w_this * ((Xc[curr_pos] - m) * (Xc[curr_pos] - m_prev));
+                m = std::fma(w_this, (Xc[curr_pos] - m) / added, m);
+                s = std::fma(w_this, (Xc[curr_pos] - m) * (Xc[curr_pos] - m_prev), s);
                 m_prev = m;
             }
 
@@ -412,12 +412,12 @@ void calc_mean_and_sd_weighted(size_t *restrict ix_arr, size_t st, size_t end, s
        The difference can be put to a closed form. */
     if (cnt > added)
     {
-        s += square(m) * (added * ((long double)1 - added/cnt));
+        s += square(m) * (added * ((long double)1 - (long double)added/(long double)cnt));
         m *= added / cnt;
     }
 
     x_mean = m;
-    x_sd   = std::sqrt(s / cnt);
+    x_sd   = std::sqrt(s / (long double)cnt);
 }
 
 template <class real_t_, class sparse_ix, class mapping>
@@ -434,12 +434,12 @@ double calc_mean_only_weighted(size_t *restrict ix_arr, size_t st, size_t end, s
     size_t ind_end_col = (size_t) Xc_ind[end_col];
     size_t *ptr_st = std::lower_bound(ix_arr + st, ix_arr + end + 1, (size_t)Xc_ind[st_col]);
 
-    long double cnt = 0.;
+    ldouble_safe cnt = 0.;
     for (size_t row = st; row <= end; row++)
         cnt += w[ix_arr[row]];
-    long double added = 0;
-    double m = 0;
-    double w_this;
+    ldouble_safe added = 0;
+    ldouble_safe m = 0;
+    ldouble_safe w_this;
 
     for (size_t *row = ptr_st;
          row != ix_arr + end + 1 && curr_pos != end_col + 1 && ind_end_col >= *row;
@@ -474,7 +474,7 @@ double calc_mean_only_weighted(size_t *restrict ix_arr, size_t st, size_t end, s
         return 0;
 
     if (cnt > added)
-        m *= added / cnt;
+        m *= (long double)added / (long double)cnt;
 
     return m;
 }
@@ -567,7 +567,7 @@ void add_linear_comb_weighted(size_t ix_arr[], size_t st, size_t end, double *re
     size_t cnt = 0;
     size_t cnt_NA = 0;
     double *restrict res_write = res - st;
-    long double cumw = 0;
+    ldouble_safe cumw = 0;
     double w_this;
     /* TODO: these buffers should be allocated externally */
     std::vector<double> obs_weight;
@@ -616,12 +616,12 @@ void add_linear_comb_weighted(size_t ix_arr[], size_t st, size_t end, double *re
         }
 
 
-        long double mid_point = cumw / 2.;
+        long double mid_point = cumw / 2.0l;
         std::vector<size_t> sorted_ix(cnt);
         std::iota(sorted_ix.begin(), sorted_ix.end(), (size_t)0);
         std::sort(sorted_ix.begin(), sorted_ix.end(),
                   [&buffer_arr](const size_t a, const size_t b){return buffer_arr[a] < buffer_arr[b];});
-        long double currw = 0;
+        ldouble_safe currw = 0;
         fill_val = buffer_arr[sorted_ix.back()]; /* <- will overwrite later */
         /* TODO: is this median calculation correct? should it do a weighted interpolation? */
         for (size_t ix = 0; ix < cnt; ix++)
@@ -894,13 +894,13 @@ void add_linear_comb_weighted(size_t *restrict ix_arr, size_t st, size_t end, si
             }
         }
 
-        long double cumw = std::accumulate(obs_weight.begin(), obs_weight.begin() + end_new, (long double)0);
-        long double mid_point = cumw / 2.0l;
+        ldouble_safe cumw = std::accumulate(obs_weight.begin(), obs_weight.begin() + end_new, (ldouble_safe)0);
+        ldouble_safe mid_point = cumw / 2.0l;
         std::vector<size_t> sorted_ix(end_new);
         std::iota(sorted_ix.begin(), sorted_ix.end(), (size_t)0);
         std::sort(sorted_ix.begin(), sorted_ix.end(),
                   [&denseX](const size_t a, const size_t b){return denseX[a] < denseX[b];});
-        long double currw = 0;
+        ldouble_safe currw = 0;
         fill_val = denseX[sorted_ix.back()]; /* <- will overwrite later */
         /* TODO: is this median calculation correct? should it do a weighted interpolation? */
         for (size_t ix = 0; ix < end_new; ix++)
@@ -1228,7 +1228,7 @@ void add_linear_comb_weighted(size_t *restrict ix_arr, size_t st, size_t end, do
             }
 
             /* TODO: this buffer should be allocated externally */
-            std::vector<long double> buffer_cnt(ncat+1, 0.);
+            std::vector<ldouble_safe> buffer_cnt(ncat+1, 0.);
             switch(missing_action)
             {
                 case Fail:
@@ -1265,7 +1265,7 @@ void add_linear_comb_weighted(size_t *restrict ix_arr, size_t st, size_t end, do
             {
                 case Smallest:
                 {
-                    long double smallest = HUGE_VALL;
+                    ldouble_safe smallest = std::numeric_limits<ldouble_safe>::infinity();
                     int cat_smallest = 0;
                     for (int cat = 0; cat < ncat; cat++)
                     {
@@ -1282,7 +1282,7 @@ void add_linear_comb_weighted(size_t *restrict ix_arr, size_t st, size_t end, do
                 default:
                 {
                     /* Determine imputation value as the category in sorted order that gives 50% + 1 */
-                    long double cnt_l = std::accumulate(buffer_cnt.begin(), buffer_cnt.begin() + ncat, (long double)0.);
+                    ldouble_safe cnt_l = std::accumulate(buffer_cnt.begin(), buffer_cnt.begin() + ncat, (ldouble_safe)0);
                     std::iota(buffer_pos, buffer_pos + ncat, (size_t)0);
                     std::sort(buffer_pos, buffer_pos + ncat, [&cat_coef](const size_t a, const size_t b){return cat_coef[a] < cat_coef[b];});
 
@@ -1290,7 +1290,7 @@ void add_linear_comb_weighted(size_t *restrict ix_arr, size_t st, size_t end, do
                     int cat;
                     for (cat = 0; cat < ncat; cat++)
                     {
-                        cumprob += (long double)buffer_cnt[buffer_pos[cat]] / cnt_l;
+                        cumprob += buffer_cnt[buffer_pos[cat]] / cnt_l;
                         if (cumprob >= .5) break;
                     }
                     // cat = std::min(cat, ncat); /* in case it picks the last one */
