@@ -18,11 +18,25 @@
 *     [5] https://sourceforge.net/projects/iforest/
 *     [6] https://math.stackexchange.com/questions/3388518/expected-number-of-paths-required-to-separate-elements-in-a-binary-tree
 *     [7] Quinlan, J. Ross. C4. 5: programs for machine learning. Elsevier, 2014.
-*     [8] Cortes, David. "Distance approximation using Isolation Forests." arXiv preprint arXiv:1910.12362 (2019).
-*     [9] Cortes, David. "Imputing missing values with unsupervised random trees." arXiv preprint arXiv:1911.06646 (2019).
+*     [8] Cortes, David.
+*         "Distance approximation using Isolation Forests."
+*         arXiv preprint arXiv:1910.12362 (2019).
+*     [9] Cortes, David.
+*         "Imputing missing values with unsupervised random trees."
+*         arXiv preprint arXiv:1911.06646 (2019).
+*     [10] https://math.stackexchange.com/questions/3333220/expected-average-depth-in-random-binary-tree-constructed-top-to-bottom
+*     [11] Cortes, David.
+*          "Revisiting randomized choices in isolation forests."
+*          arXiv preprint arXiv:2110.13402 (2021).
+*     [12] Guha, Sudipto, et al.
+*          "Robust random cut forest based anomaly detection on streams."
+*          International conference on machine learning. PMLR, 2016.
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 * 
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -93,7 +107,26 @@ std::string generate_sql_with_select_from(IsoForest *model_outputs, ExtIsoForest
                                                        categ_levels,
                                                        false, index1, false, 0,
                                                        nthreads);
-    std::string out = std::accumulate(tree_conds.begin(), tree_conds.end(), std::string("SELECT\nPOWER(2.0, -(0.0"),
+    bool is_density = (model_outputs != NULL && model_outputs->scoring_metric == Density) ||
+                      (model_outputs_ext != NULL && model_outputs_ext->scoring_metric == Density);
+    bool is_bdens   = (model_outputs != NULL && model_outputs->scoring_metric == BoxedDensity) ||
+                      (model_outputs_ext != NULL && model_outputs_ext->scoring_metric == BoxedDensity);
+    bool is_bdens2  = (model_outputs != NULL && model_outputs->scoring_metric == BoxedDensity) ||
+                      (model_outputs_ext != NULL && model_outputs_ext->scoring_metric == BoxedDensity);
+    bool is_bratio  = (model_outputs != NULL && model_outputs->scoring_metric == BoxedRatio) ||
+                      (model_outputs_ext != NULL && model_outputs_ext->scoring_metric == BoxedRatio);
+    is_density = is_density || is_bdens2;
+    std::string out = std::accumulate(tree_conds.begin(), tree_conds.end(),
+                                      is_density?
+                                          std::string("SELECT\n(-(0.0")
+                                          :
+                                          (is_bdens?
+                                               std::string("SELECT\n((0.0")
+                                               :
+                                               (is_bratio?
+                                                    std::string("SELECT\n((0.0")
+                                                    :
+                                                    std::string("SELECT\nPOWER(2.0, -(0.0"))),
                                       [&tree_conds, &index1](std::string &a, std::string &b)
                                       {return a
                                                 + std::string(" + \n---BEGIN TREE ")
@@ -174,7 +207,7 @@ std::vector<std::string> generate_sql(IsoForest *model_outputs, ExtIsoForest *mo
     size_t_for loop_end = ntrees_use;
     if (single_tree)
     {
-        loop_st = tree_num;
+        loop_st = tree_num - index1;
         loop_end = loop_st + 1;
     }
 
@@ -228,7 +261,8 @@ std::vector<std::string> generate_sql(IsoForest *model_outputs, ExtIsoForest *mo
                 (model_outputs_ext == NULL)? (NULL) : &(model_outputs_ext->hplanes[tree]),
                 output_score,
                 0, index1, initial_str, all_node_rules[single_tree? 0 : tree],
-                conditions_left, conditions_right
+                conditions_left, conditions_right,
+                model_outputs, model_outputs_ext
             );
 
             /* Code below doesn't compile with MSVC (stuck with an OMP standard that's >20 years old) */
@@ -279,16 +313,23 @@ std::vector<std::string> generate_sql(IsoForest *model_outputs, ExtIsoForest *mo
 
 void generate_tree_rules(std::vector<IsoTree> *trees, std::vector<IsoHPlane> *hplanes, bool output_score,
                          size_t curr_ix, bool index1, std::string &prev_cond, std::vector<std::string> &node_rules,
-                         std::vector<std::string> &conditions_left, std::vector<std::string> &conditions_right)
+                         std::vector<std::string> &conditions_left, std::vector<std::string> &conditions_right,
+                         const IsoForest *model_outputs, const ExtIsoForest *model_outputs_ext)
 {
-    if ((trees != NULL && (*trees)[curr_ix].score >= 0) ||
-        (hplanes != NULL && (*hplanes)[curr_ix].score >= 0))
+    // if ((trees != NULL && (*trees)[curr_ix].score >= 0) ||
+    //     (hplanes != NULL && (*hplanes)[curr_ix].score >= 0))
+    if ((trees != NULL && (*trees)[curr_ix].tree_left == 0) ||
+        (hplanes != NULL && (*hplanes)[curr_ix].hplane_left == 0))
     {
         node_rules.push_back(prev_cond
                                 + std::string("\tTHEN ")
                                 + (output_score?
                                     (std::to_string((trees != NULL)?
-                                        ((*trees)[curr_ix].score) : ((*hplanes)[curr_ix].score)))
+                                        ((model_outputs->scoring_metric != Density && model_outputs->scoring_metric != BoxedRatio)?
+                                            (*trees)[curr_ix].score : (-(*trees)[curr_ix].score))
+                                            :
+                                        ((model_outputs_ext->scoring_metric != Density && model_outputs_ext->scoring_metric != BoxedRatio)?
+                                            (*hplanes)[curr_ix].score : (-(*hplanes)[curr_ix].score))))
                                         :
                                     (std::to_string(node_rules.size() + (size_t)index1)))
                                 + std::string("\n---end of terminal node ")
@@ -306,7 +347,7 @@ void generate_tree_rules(std::vector<IsoTree> *trees, std::vector<IsoHPlane> *hp
                         (trees != NULL)?
                             ((*trees)[curr_ix].tree_left) : ((*hplanes)[curr_ix].hplane_left),
                         index1, cond_left, node_rules,
-                        conditions_left, conditions_right);
+                        conditions_left, conditions_right, model_outputs, model_outputs_ext);
     cond_left.clear();
     std::string cond_right = prev_cond
                                 + ((curr_ix > 0)? std::string("\t\tAND (") : std::string("\t\t    ("))
@@ -316,7 +357,7 @@ void generate_tree_rules(std::vector<IsoTree> *trees, std::vector<IsoHPlane> *hp
                         (trees != NULL)?
                             ((*trees)[curr_ix].tree_right) : ((*hplanes)[curr_ix].hplane_right),
                         index1, cond_right, node_rules,
-                        conditions_left, conditions_right);
+                        conditions_left, conditions_right, model_outputs, model_outputs_ext);
 }
 
 
@@ -327,7 +368,8 @@ void extract_cond_isotree(IsoForest &model, IsoTree &tree,
 {
     cond_left = std::string("");
     cond_right = std::string("");
-    if (tree.score >= 0.)
+    // if (tree.score >= 0.)
+    if (tree.tree_left == 0)
         return;
 
     switch(tree.col_type)
@@ -496,7 +538,8 @@ void extract_cond_ext_isotree(ExtIsoForest &model, IsoHPlane &hplane,
 {
     cond_left = std::string("");
     cond_right = std::string("");
-    if (hplane.score >= 0.)
+    // if (hplane.score >= 0.)
+    if (hplane.hplane_left == 0)
         return;
 
     std::string hplane_conds = std::string("");
